#include "StdAfx.h"
#include <shlobj.h>
#include <time.h>
#include "CredentialCache.h"
#include "RegSettings.h"
#include "Crypto.h"
#include "Logging.h"

using namespace std;

static LPCTSTR APPDATA_FOLDER = _T("\\PINsafeGINA");
static LPCTSTR CACHE_FILE_TEMPLATE = _T("%s%sPINsafe_%s.cache");
static list <CredentialCache*> g_CacheTable;
static LPTSTR g_ApplicationDataFolder = NULL;
static LPTSTR g_Secret = NULL;

static LPCSTR alpha6 = "+-0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz";

LPSTR ByteArrayTo6BitString(LPBYTE hash, DWORD hashLen)
{
	DWORD strLen = (hashLen*8 + 5)/6;
	LPSTR result = new char[strLen+1];
	memset(result, 0, strLen+1);

	for (int idx=0, ndx=0; ndx<hashLen; idx++) {
		for (int maskL = 0x3f, maskR = 0x30, shift = 2; (maskL > 0) && (ndx < hashLen); shift += 2) {
			BYTE x = hash[ndx++];
			result[idx++] |= (x >> shift) & maskL;
			result[idx] = (BYTE)((x << (6-shift)) & maskR);
			maskL >>= 2;
			maskR = 0x30 | (maskR >> 2);
		}
	}
	// Now convert indices to characters.
	for (int idx=0; idx<strLen; idx++) {
		result[idx] = alpha6[result[idx]];
	}

	return result;
}

LPCTSTR HashCredentials(LPCTSTR username, LPCTSTR secString, LPCTSTR password, LPCTSTR pin, LPCTSTR secret)
{
	size_t strLen = _tcslen(username) + _tcslen(secString) + _tcslen(password) + _tcslen(pin) + _tcslen(secret);
	LPSTR hashString = new char[strLen+1];
#ifdef UNICODE
	LPWSTR wHashString = new WCHAR[strLen+1];
	_tcscpy(wHashString, secret);
	_tcscat(wHashString, username);
	_tcslwr(wHashString + _tcslen(secret)); // username must be in lower case
	_tcscat(wHashString, secString);
	_tcscat(wHashString, password);
	_tcscat(wHashString, pin);
	WideCharToMultiByte(CP_ACP, 0, wHashString, -1, hashString, strLen+1, NULL, NULL);
	delete wHashString;
#else
	_tcscpy(hashString, secret);
	_tcscat(hashString, username);
	_tcslwr(hashString + _tcslen(secret)); // username must be in lower case
	_tcscat(hashString, secString);
	_tcscat(hashString, password);
	_tcscat(hashString, pin);
#endif
	CCrypto crypt;
	crypt.SessionStart();
	DWORD hashLen;
	BYTE* hash = crypt.Hash((BYTE*)hashString, strLen, &hashLen, CALG_SHA1);
	LPSTR result = ByteArrayTo6BitString(hash, hashLen);
	delete hash;
	crypt.SessionEnd();
#ifdef UNICODE
	size_t resLen = strlen(result)+1;
	LPWSTR wResult = new wchar_t[resLen];
	MultiByteToWideChar(CP_ACP, 0, result, -1, wResult, resLen);
	delete result;
	return wResult;
#else
	return result;
#endif
}

LPTSTR PinFromOTC(LPCTSTR secString, LPCTSTR otc)
{
	size_t pinLen = _tcslen(otc);
	LPTSTR pin = new TCHAR[pinLen+1];
	for (int idx=0; idx<pinLen; idx++) {
		LPCTSTR cdx = _tcschr(secString, otc[idx]);
		if (!cdx) {
			return FALSE;
		}
		TCHAR ndx = (TCHAR)(cdx - secString);
		pin[idx] = (ndx == 9) ? '0' : (ndx + '1');
	}
	pin[pinLen] = 0;

	return pin;
}

LPTSTR CreateNewSecurityString(void)
{
	int placed = 0;
	int mask[10] = {0,0,0,0,0,0,0,0,0,0};
	unsigned int digit;
	
	LPTSTR securityString = new TCHAR[11];

	while (placed < 10)
	{
		do {
			rand_s(&digit);
			digit &= 0xf;
		} while (digit > 9);

		if (mask[digit] == 0)
		{
			mask[digit] = 1;

			securityString[placed] = ('0' + digit);

			placed++;
		}
	}
	securityString[10] = 0;

	return securityString;
}

BOOL CredentialCache::LocalUserExists(LPCTSTR username)
{
	LPTSTR cacheFile = CacheFileForUser(username);
	BOOL result = PathFileExists(cacheFile);
	delete cacheFile;
	return result;
}

CredentialCache* CredentialCache::GetCacheForUser(LPCTSTR username)
{
	LPTSTR lUsername = new TCHAR[_tcslen(username)+1];
	_tcscpy(lUsername, username);
	_tcslwr(lUsername);
	list <CredentialCache*>::iterator iter = g_CacheTable.begin();
	do {
		if (iter == g_CacheTable.end()) {
			// Create a new entry
			CredentialCache* newCache = new CredentialCache(lUsername);
			g_CacheTable.push_back(newCache);
			return newCache;
		}
		if (_tcscmp(lUsername, (*iter)->m_Username) == 0) {
			break;
		}
		iter ++;
	} while (true);
	return *iter;
}

CredentialCache::CredentialCache(LPCTSTR username)
{
	m_Username = new TCHAR[_tcslen(username)+1];
	_tcscpy(m_Username, username);
	_tcslwr(m_Username);
	m_Pin = NULL;
	m_Password = NULL;
	m_CacheFile = NULL;
	RegSettings* settings = RegSettings::GetSettings();
	lockoutCount = settings->GetLockoutCount();
	failureCount = (lockoutCount > 0) ? 0 : -1;

	// Load the cached credentials for the given user.
	LoadStrings();
}

void CredentialCache::CheckCacheFolder()
{
	if (g_ApplicationDataFolder == NULL) {
		g_ApplicationDataFolder = new TCHAR[MAX_PATH];
		SHGetFolderPath(NULL, CSIDL_COMMON_APPDATA, NULL, 0, g_ApplicationDataFolder);
		_tcscat(g_ApplicationDataFolder, APPDATA_FOLDER);
		// Create the folder, in case it doesn't already exist
		CreateDirectory(g_ApplicationDataFolder, NULL);
	}
}

LPTSTR CredentialCache::CacheFileForUser(LPCTSTR username)
{
	CheckCacheFolder();
	LPTSTR cacheFile = new TCHAR[_tcslen(g_ApplicationDataFolder) + _tcslen(APPDATA_FOLDER) + _tcslen(CACHE_FILE_TEMPLATE) + _tcslen(username)];
	_stprintf(cacheFile, CACHE_FILE_TEMPLATE, g_ApplicationDataFolder, APPDATA_FOLDER, username);
	return cacheFile;
}

void CredentialCache::LoadStrings()
{
	if (m_CacheFile == NULL) {
		m_CacheFile = CacheFileForUser(m_Username);
	}
	m_CachedStrings.clear();
	m_UsedStrings.clear();
	FILE* hFile = _tfopen(m_CacheFile, _T("r+b"));
	TCHAR stringAndHash[0x200];
	if (hFile != NULL) {
		// Grab the user's lockout count first
		int tmpLockout;
		if (fread((void*)&tmpLockout, sizeof(int), 1, hFile)) {
			if (lockoutCount > 0) {
				if (tmpLockout > 0) {
					failureCount = tmpLockout;
				} else {
					failureCount = 0;
				}
			} else {
				failureCount = -1;
			}
		}
		size_t result;
		PCacheString newEntry = new CacheString();
		do {
			DWORD offset = ftell(hFile);
			result = fread(newEntry, sizeof(CacheString), 1, hFile);
			if (result > 0) {
				if (!newEntry->deleted) {
					newEntry->offset = offset;
					if (newEntry->used) {
						m_UsedStrings.push_back(newEntry);
					} else {
						m_CachedStrings.push_back(newEntry);
					}
					newEntry = new CacheString();
				}
			}
		} while (!feof(hFile));
		// The last entry will not have been used.
		delete newEntry;
		if (m_CachedStrings.size() > 0) {
			// We have unused strings, so clear out the used ones.
			for (list<PCacheString>::iterator entry = m_UsedStrings.begin(); entry != m_UsedStrings.end(); ) {
				newEntry = *entry;
				newEntry->deleted = TRUE;
				fseek(hFile, newEntry->offset, SEEK_SET);
				fwrite(newEntry, sizeof(CacheString), 1, hFile);
				m_UsedStrings.erase(entry);
				entry = m_UsedStrings.begin();
			}
		} else {
			// No unused strings, so re-use the old ones.
			for (list<PCacheString>::iterator entry = m_UsedStrings.begin(); entry != m_UsedStrings.end(); ) {
				newEntry = *entry;
				newEntry->used = 0;
				fseek(hFile, newEntry->offset, SEEK_SET);
				fwrite(newEntry, sizeof(CacheString), 1, hFile);
				m_CachedStrings.push_back(newEntry);
				m_UsedStrings.erase(entry);
				entry = m_UsedStrings.begin();
			}
		}
		fclose(hFile);
	}
}

CredentialCache::~CredentialCache(void)
{
	if (m_CacheFile != NULL) {
		delete m_CacheFile;
	}
	delete m_Username;
	if (m_Pin != NULL) {
		delete m_Pin;
	}
	for (list<PCacheString>::iterator entry = m_CachedStrings.begin(); entry != m_CachedStrings.end(); ) {
		delete (*entry);
		m_CachedStrings.erase(entry);
		entry = m_CachedStrings.begin();
	}
	for (list<PCacheString>::iterator entry = m_UsedStrings.begin(); entry != m_UsedStrings.end(); ) {
		delete (*entry);
		m_UsedStrings.erase(entry);
		entry = m_UsedStrings.begin();
	}
	g_CacheTable.remove(this);
}

LPCTSTR CredentialCache::GetSecurityString(BOOL getNew)
{
	LPCTSTR newString = NULL;

	if (m_CacheFile == NULL) {
		newString = CreateNewSecurityString();
	} else if (getNew) {
		FILE *hFile = _tfopen(m_CacheFile, _T("r+b"));
		if (hFile == NULL) {
			LogError(L"Unable to open strings cache for %s", m_Username);
			newString = CreateNewSecurityString();
		} else {
			RegSettings* settings = RegSettings::GetSettings();
			// Take a string from the cached strings and move it to the used strings.
			if (m_CachedStrings.size() > 0) {
				list<PCacheString>::const_iterator entry = m_CachedStrings.begin();
				PCacheString cacheString = *entry;
				newString = cacheString->securityString;
				m_UsedStrings.push_front(cacheString);
				// Update the file and mark this entry as used.
				cacheString->used = time(NULL);
				fseek(hFile, cacheString->offset, SEEK_SET);
				fwrite(cacheString, sizeof(CacheString), 1, hFile);
				m_CachedStrings.pop_front();
				// While we're in here, let's expire all the old strings.
				DWORD expireTime = time(NULL) - settings->GetSessionTimeout();
				for (list<PCacheString>::iterator entry = m_UsedStrings.begin(); entry != m_UsedStrings.end(); ) {
					PCacheString s = *entry;
					if (s->used < expireTime) {
						s->deleted = TRUE;
						fseek(hFile, s->offset, SEEK_SET);
						fwrite(s, sizeof(CacheString), 1, hFile);
						entry = m_UsedStrings.erase(entry);
					} else {
						entry ++;
					}
				}
			} else if (m_UsedStrings.size() > 0) {
				// OK, if we don't have any unused ones left, re-use the oldest one.
				list<PCacheString>::reverse_iterator entry = m_UsedStrings.rbegin();
				PCacheString newEntry = *entry;
				newString = newEntry->securityString;
				newEntry->used = time(NULL);
				if (m_UsedStrings.size() > 1) {
					// Move this string to the front of the list
					m_UsedStrings.pop_back();
					m_UsedStrings.push_front(newEntry);
				}
			}
			DWORD cacheLowCount = settings->GetCacheLowCount();
			DWORD cacheUpdateCount = settings->GetCachedStringCount();
			// If we have fewer than the require number left, create several new ones now, rather than on demand.
			if (m_CachedStrings.size() <= cacheLowCount && g_Secret != NULL && m_Pin != NULL) {
				for (int idx=0; idx<cacheUpdateCount; idx++) {
					// Generate a new one
					LPCTSTR nextString = CreateNewSecurityString();
					LPCTSTR hashString = HashCredentials(m_Username, nextString, m_Password, m_Pin, g_Secret);
					PCacheString newEntry = new CacheString();
					_tcscpy(newEntry->securityString, nextString);
					_tcscpy(newEntry->hash, hashString);
					newEntry->used = 0;
					newEntry->deleted = 0;
					m_CachedStrings.push_back(newEntry);
					delete nextString;
					delete hashString;
				}
				// Insert these entries as new, used entries in the file
				CacheString tmpString;
				fseek(hFile, 0, SEEK_SET);
				DWORD offset = 0;
				list<PCacheString>::const_iterator entry = m_CachedStrings.begin();
				while (entry != m_CachedStrings.end()) {
					if (!fread(&tmpString, sizeof(CacheString), 1, hFile)) {
						break;
					}
					if (tmpString.deleted) {
						fseek(hFile, offset, SEEK_SET);
						(*entry)->offset = offset;
						fwrite(*entry, sizeof(CacheString), 1, hFile);
						entry ++;
						fseek(hFile, 0, SEEK_CUR);
					}
					offset += sizeof(CacheString);
				}
				fseek(hFile, 0, SEEK_END);
				offset = ftell(hFile);
				while (entry != m_CachedStrings.end()) {
					(*entry)->offset = offset;
					fwrite((*entry), sizeof(CacheString), 1, hFile);
					offset += sizeof(CacheString);
					entry ++;
				}
			}
			fclose(hFile);
		}
	} else if (m_UsedStrings.size() > 0) {
		list<PCacheString>::iterator entry = m_UsedStrings.begin();
		newString = (*entry)->securityString;
	} else {
		// No existing string and not allowed to create a new one, so just create one at random.
		newString = CreateNewSecurityString();
	}
	return newString;
}

BOOL CredentialCache::Authenticate(LPCTSTR secString, LPCTSTR password, LPCTSTR otc)
{
	if (g_Secret == NULL) {
		RegSettings *settings = RegSettings::GetSettings();
		LPCWSTR secret = settings->GetSecret();
		DWORD secLen = wcslen(secret)+1;
		g_Secret = new TCHAR[secLen];
#ifdef UNICODE
		wcscpy(g_Secret, secret);
#else
		WideCharToMultiByte(CP_ACP, 0, secret, -1, g_Secret, secLen, NULL, NULL);
#endif
	}
	// First of all, check if they have used up their retries
	if (lockoutCount > 0 && failureCount >= lockoutCount) {
		return FALSE;
	}
	list<PCacheString>::iterator entry = m_UsedStrings.begin();
	do {
		if (entry == m_UsedStrings.end()) {
			// No such security string.
			return FALSE;
		}
		if (_tcsicmp(secString, (*entry)->securityString) == 0) {
			break;
		}
		entry ++;
	} while (true);
	// This string can't be re-used
	PCacheString string = *entry;
	m_UsedStrings.erase(entry);
	string->deleted = TRUE;
	// Update the file and mark this entry as deleted.
	FILE *hFile = _tfopen(m_CacheFile, _T("r+b"));
	fseek(hFile, string->offset, SEEK_SET);
	fwrite(string, sizeof(CacheString), 1, hFile);
	fclose(hFile);
	LPCTSTR expectedHash = string->hash;
	LPTSTR pin = PinFromOTC(secString, otc);
	LPCTSTR newHash = HashCredentials(m_Username, secString, password, pin, g_Secret);
	BOOL result = (_tcscmp(expectedHash, newHash) == 0);
	BOOL storeFailure = FALSE;
	if (result) {
		// save the pin and password.
		if (m_Pin != NULL) {
			if (_tcscmp(m_Pin, pin) != 0) {
				// PIN has changed.
				delete m_Pin;
				m_Pin = pin;
			} else {
				delete pin;
			}
		} else {
			m_Pin = pin;
		}
		if (m_Password != NULL) {
			if (_tcscmp(m_Password, password) != 0) {
				// Password has changed.
				delete m_Password;
				m_Password = new TCHAR[_tcslen(password)+1];
				_tcscpy(m_Password, password);
			}
		} else {
			m_Password = new TCHAR[_tcslen(password)+1];
			_tcscpy(m_Password, password);
		}
		// Reset the failure count
		if (lockoutCount > 0) {
			storeFailure = (failureCount > 0);
			failureCount = 0;
		}
	} else {
		if (lockoutCount > 0) {
			storeFailure = TRUE;
			failureCount ++;
		}
		delete pin;
	}
	// Store the updated failure count.
	if (storeFailure) {
		FILE* hFile = _tfopen(m_CacheFile, _T("r+b"));
		fseek(hFile, 0, SEEK_SET);
		fwrite(&failureCount, sizeof(int), 1, hFile);
		fclose(hFile);
	}
	delete newHash;
	return result;
}

void CredentialCache::StoreCredentials(PCacheString* newStrings, DWORD cString)
{
	if (m_CacheFile == NULL) {
		m_CacheFile = new TCHAR[_tcslen(g_ApplicationDataFolder) + _tcslen(APPDATA_FOLDER) + _tcslen(CACHE_FILE_TEMPLATE) + _tcslen(m_Username)];
		_stprintf(m_CacheFile, CACHE_FILE_TEMPLATE, g_ApplicationDataFolder, APPDATA_FOLDER, m_Username);
	}
	// We assume if we're storing new credentials that the user must have successfully authenticated to PINsafe.
	if (lockoutCount > 0) {
		failureCount = 0;
	}
	FILE* hFile = _tfopen(m_CacheFile, _T("wb"));
	if (hFile) {
		// Write the failure count.
		fwrite((void*)(&failureCount), sizeof(int), 1, hFile);
		m_CachedStrings.clear();
		m_UsedStrings.clear();
		// Write the security strings.
		DWORD offset = ftell(hFile);
		for (int idx=0; idx<cString; idx++) {
			newStrings[idx]->offset = offset;
			fwrite(newStrings[idx], sizeof(CacheString), 1, hFile);
			offset += sizeof(CacheString);
			m_CachedStrings.push_back(newStrings[idx]);
		}
	}
	fclose(hFile);
}
