#include "stdafx.h"

#include "PINsafeConnect.h"
#include "HttpSocket.h"
#include "CredentialCache.h"
#include "logging.h"

// Constant XML strings for agent requests.

LPCWSTR SAS_REQUEST_TAG = L"SASRequest";
LPCWSTR ACTION_TAG = L"Action";
LPCWSTR ACTION_PING = L"ping";
LPCWSTR ACTION_STARTSESSION = L"sessionstart";
LPCWSTR ACTION_LOGIN = L"login";
LPCWSTR ACTION_CHANGEPIN = L"changepin";
LPCWSTR ACTION_EXISTS = L"exists";
LPCWSTR SECRET_TAG = L"Secret";
LPCWSTR THIRD_PARTY_TAG = L"ThirdParty";
LPCWSTR NAME_TAG = L"Name";
LPCWSTR DATA_TAG = L"Data";
LPCWSTR USERNAME_TAG = L"Username";
LPCWSTR PASSWORD_TAG = L"Password";
LPCWSTR OTC_TAG = L"OTC";
LPCWSTR GINA_NAME_TAG = L"Name";
LPCWSTR GINA_NAME_VALUE = L"WindowsGINA";
LPCWSTR GINA_DATA_TAG = L"Data";
LPCWSTR RESULT_TAG = L"Result";
LPCWSTR RESULT_PASS = L"PASS";
LPCWSTR RESULT_FAIL = L"FAIL";
LPCWSTR WARNING_TAG = L"Warning";
LPCWSTR SESSION_ID_TAG = L"SessionID";
LPCWSTR NEWPASSWORD_TAG = L"NewPassword";
LPCWSTR NEWOTC_TAG = L"NewOTC";
LPCWSTR AGENT_URL = L"AgentXML";
LPCWSTR IMAGE_URL = L"SCImage";
LPCWSTR MESSAGE_URL = L"DCMessage";
LPCWSTR HTTP_PROT = L"http";
LPCWSTR HTTP_SECURE_PROT = L"https";
LPCWSTR SESSION_ARG = L"?sessionid=";
LPCWSTR USERNAME_ARG = L"?username=";
LPCWSTR PIN_EXPIRES_WARNING = L"AGENT_WARN_PIN_EXPIRY";
LPCWSTR PIN_CHANGE_REQUIRED = L"AGENT_WARN_CHANGE_PIN";
LPCWSTR SEC_STRINGS_TAG = L"SecurityStrings";
LPCWSTR SEC_STRING_TAG = L"SecurityString";
LPCWSTR STRING_TAG = L"String";
LPCWSTR HASH_TAG = L"Hash";

// Utility functions for building an XML string.

// Append a start tag to an XML string
LPWSTR AppendStartTag(LPWSTR buffer, LPCWSTR element, int &charsLeft) {
	int len = wcslen(element) + 2;
	if (charsLeft <= len) {
		// Not enough room left.
		return NULL;
	}
	swprintf(buffer, len+1, L"<%s>", element);
	charsLeft -= len;
	return buffer + len;
}

// Append an end tag to an XML string
LPWSTR AppendEndTag(LPWSTR buffer, LPCWSTR element, int &charsLeft) {
	int len = wcslen(element) + 3;
	if (charsLeft <= len) {
		// Not enough room left.
		return NULL;
	}
	swprintf(buffer, len+1, L"</%s>", element);
	charsLeft -= len;
	return buffer + len;
}

// Append a value to a buffer, applying XML escaping as required.
// Parameters:
//   buffer - The buffer to be appended to.
//   value - The value to append.
//   charsLeft - On input, the number of characters available. 
//               On output, the number of characters left.
// Returns a pointer to the next available character location, or NULL if not enough space.
LPWSTR EscapeXmlValue(LPWSTR buffer, LPCWSTR value, int &charsLeft) {
	LPWSTR bufPtr;
	LPCWSTR valPtr;
	int myCharsLeft = charsLeft;
	for (valPtr = value, bufPtr = buffer; *valPtr != 0; valPtr ++) {
		if (myCharsLeft <= 0) {
			*buffer = 0; // Make sure the old buffer is terminated properly
			// charsLeft is not altered.
			return NULL;
		}
		switch (*valPtr) {
			case L'<': {
				if (myCharsLeft < 4) {
					// Not enough room
					*buffer = 0 ;
					return NULL;
				}
				wcscpy(bufPtr, L"&lt;");
				bufPtr += 4;
				myCharsLeft -= 4;
				break;
			}
			case L'>': {
				if (myCharsLeft < 4) {
					// Not enough room
					*buffer = 0 ;
					return NULL;
				}
				wcscpy(bufPtr, L"&gt;");
				bufPtr += 4;
				myCharsLeft -= 4;
				break;
			}
			case L'&': {
				if (myCharsLeft < 5) {
					// Not enough room
					*buffer = 0 ;
					return NULL;
				}
				wcscpy(bufPtr, L"&amp;");
				bufPtr += 4;
				myCharsLeft -= 4;
				break;
			}
			default: {
				if (iswprint(*valPtr)) {
					*bufPtr++ = *valPtr;
					myCharsLeft --;
				} else if (*valPtr < 0x100) {
					// Use the 2-digit hex code.
					if (myCharsLeft < 6) {
						*buffer = 0;
						return NULL;
					}
					swprintf(bufPtr, 6, L"&#x%02x;", *valPtr);
					*bufPtr += 6;
					myCharsLeft -= 6;
				} else {
					// Use the 4-digit hex code.
					if (myCharsLeft < 8) {
						*buffer = 0;
						return NULL;
					}
					swprintf(bufPtr, 8, L"&#x%04x;", *valPtr);
					*bufPtr += 8;
					myCharsLeft -= 8;
				}
				break;
			}
		}
	}
	// On successful exit, update charsLeft.
	charsLeft = myCharsLeft;
	*bufPtr = 0;
	return bufPtr;
}

// Append a value to an XML string, surrounded by start and end tags
LPWSTR AppendTagAndValue(LPWSTR buffer, LPCWSTR element, LPCWSTR value, int &charsLeft) {
	int myCharsLeft = charsLeft;
	LPWSTR bufPtr = AppendStartTag(buffer, element, myCharsLeft);
	if (bufPtr != NULL) {
		bufPtr = EscapeXmlValue(bufPtr, value, myCharsLeft);
		if (bufPtr != NULL) {
			bufPtr = AppendEndTag(bufPtr, element, myCharsLeft);
		}
	}
	if (bufPtr != NULL) {
		charsLeft = myCharsLeft;
		return bufPtr;
	} else {
		return NULL;
	}
}

// Extract the text content of an XML element.
// Parameters:
//   xmlBuffer	- [in] the text buffer containing the XML string
//   element	- [in] the name of the element to extract
//   value		- [out] the buffer in which to store the extracted value
//   nChars		- [in] the maximum length of the value buffer. [out] the actual length of the value buffer
// Returns a pointer to the next character after the closing element tag. NULL if the string was not found.
LPCWSTR ExtractXmlValue(LPCWSTR xmlBuffer, LPCWSTR element, LPWSTR value, int* nChars) {
	int elemLen = wcslen(element);
	LPWSTR startTag = new wchar_t[elemLen+3];
	swprintf(startTag, elemLen+3, L"<%s>", element);
	LPWSTR endTag = new wchar_t[elemLen+4];
	swprintf(endTag, elemLen+4, L"</%s>", element);
	LPCWSTR ptr1 = wcsstr(xmlBuffer, startTag);
	if (ptr1 == NULL) {
		*nChars = 0;
		return NULL;
	}
	ptr1 += elemLen + 2;
	LPCWSTR ptr2 = wcsstr(ptr1, endTag);
	if (ptr2 == NULL) {
		*nChars = 0;
		return NULL;
	}
	int maxChars = *nChars;
	int vLen = (int)(ptr2-ptr1);
	if (vLen < maxChars) {
		wcsncpy(value, ptr1, vLen);
		value[vLen] = 0;
	}
	*nChars = vLen;
	return ptr2 + wcslen(endTag);
}

// Build the PINsafe URL from its component parts.
int CreatePINsafeUrl(LPWSTR url, int maxChars, LPCWSTR server, LPCWSTR port, LPCWSTR context, BOOL secure) {
	const int bufLen = 0x200;
	wchar_t buffer[bufLen];
	LPWSTR contextBuf = new wchar_t[wcslen(context)+3];
	LPWSTR ptr = contextBuf;
	if (wcslen(context) == 0) {
		wcscpy(contextBuf, L"/");
	} else {
		if (context[0] != L'/') {
			contextBuf[0] = L'/';
			ptr ++;
		}
		wcscpy(ptr, context);
		if (wcslen(context) > 1 && context[wcslen(context)-1] != L'/') {
			wcscat(ptr, L"/");
		}
	}

	swprintf(buffer, bufLen, L"%s://%s:%s%s", (secure ? HTTP_SECURE_PROT : HTTP_PROT), server, port, contextBuf);
	delete contextBuf;
	int len = wcslen(buffer);
	if (len < maxChars) {
		wcscpy(url, buffer);
	}
	return len;
}

// Constructor
PINsafeConnect::PINsafeConnect(void)
{
	RegSettings* settings = RegSettings::GetSettings();
	GetSettings(settings);
}

// Destructor
PINsafeConnect::~PINsafeConnect(void)
{
	delete agentUrl;
	delete scImageUrl;
	delete dcImageUrl;
	delete secret;
}

// Initialise the connection parameters.
void PINsafeConnect::GetSettings(RegSettings* settings) {
	LPCWSTR regSecret = settings->GetSecret();
	secret = new wchar_t[wcslen(regSecret)+1];
	wcscpy(secret, regSecret);
	secure = settings->IsSecure();
	loginRemote = settings->LoginRemote();
	showGetImage = settings->ShowGetImage();
	allowDefaultLogin = settings->AllowDefaultLogin();
	allowSelfCert = settings->AllowSelfCertification();
	useLocal = settings->AllowLocalAuth();
	int bufSize = 0x200;
	agentUrl = new wchar_t[bufSize];
	int urlLen = CreatePINsafeUrl(agentUrl, bufSize, settings->GetServer(), settings->GetPort(), settings->GetContext(), secure);
	scImageUrl = new wchar_t[bufSize];
	dcImageUrl = new wchar_t[bufSize];
	wcscpy(scImageUrl, agentUrl);
	wcscpy(dcImageUrl, agentUrl);
	wcscat(agentUrl, AGENT_URL);
	wcscat(scImageUrl, IMAGE_URL);
	wcscat(dcImageUrl, MESSAGE_URL);
}

// Start a new session.
// Parameters:
//   username - on entry, the name of the user attempting to log in (possibly including domain).
//				on exit, the name of the user actually authenticated (without domain if appropriate).
//   sessionId - on return, the session ID.
//   maxChars - the maximum number of characters for the session ID.
// Returns true if the session was created successfully.
BOOL PINsafeConnect::StartSession(LPWSTR username, LPWSTR sessionId, int maxChars) {
	BOOL result = TryStartSession(username, sessionId, maxChars);
	if (!result) {
		LPWSTR pos = wcschr(username, L'\\');
		if (pos) {
			pos ++;
			result = TryStartSession(pos, sessionId, maxChars);
			if (result) {
				wcscpy(username, pos);
			}
		} else {
			pos = wcschr(username, L'@');
			if (pos) {
				*pos = 0;
				result = TryStartSession(username, sessionId, maxChars);
				if (!result) {
					// If this failed, we must restore the userID to what it previously was.
					*pos = L'@';
				}
			}
		}
	}
	return result;
}

BOOL PINsafeConnect::TryStartSession(LPCWSTR username, LPWSTR sessionId, int maxChars) {
	LPWSTR request = CreateStartSessionRequest(username);
	HttpSocket* socket = new HttpSocket();
	socket->SetSecure(secure);
	socket->SetAllowSelfCert(allowSelfCert);
	LPWSTR response = socket->GetPage(agentUrl, TRUE, request);
	BOOL result = FALSE;
	if (response != NULL) {
		wchar_t resBuf[0x10];
		int resLen = 0x10;
		ExtractXmlValue(response, RESULT_TAG, resBuf, &resLen);
		if (resLen > 0 && resLen < 0x10 && wcscmp(resBuf, RESULT_PASS) == 0) {
			resLen = maxChars;
			ExtractXmlValue(response, SESSION_ID_TAG, sessionId, &resLen);
			if (resLen > 0 && resLen < maxChars) {
				result = TRUE;
			}
		}
		delete response;
	}
	delete request;
	delete socket;
	return result;
}

LoginState PINsafeConnect::Authenticate(LPCWSTR username, LPCWSTR password, LPCWSTR otc, BOOL requestStrings) {
	LoginState result = LOGIN_FAILED;
	LPWSTR request = CreateLoginRequest(username, password, otc, requestStrings);
	HttpSocket* socket = new HttpSocket();
	socket->SetSecure(secure);
	socket->SetAllowSelfCert(allowSelfCert);
	LPWSTR response = socket->GetPage(agentUrl, TRUE, request);
	if (response != NULL) {
		wchar_t resBuf[0x10];
		int resLen = 0x10;
		ExtractXmlValue(response, RESULT_TAG, resBuf, &resLen);
		if (resLen > 0 && resLen < 0x10 && wcscmp(resBuf, RESULT_PASS) == 0) {
			// Get the warnings, if any
			wchar_t warnBuf[0x80];
			int warnLen = 0x80;
			ExtractXmlValue(response, WARNING_TAG, warnBuf, &warnLen);
			if (warnLen > 0) {
				if (wcscmp(warnBuf, PIN_CHANGE_REQUIRED) == 0) {
					result = LOGIN_PINCHANGE_REQUIRED;
				} else if (wcscmp(warnBuf, PIN_EXPIRES_WARNING) == 0) {
					result = LOGIN_PIN_EXPIRES_SOON;
				} else {
					result = LOGIN_PASSED;
				}
			} else {
				result = LOGIN_PASSED;
			}
		}
		if (result == LOGIN_PASSED && requestStrings) {
			// Extract the cached security strings
			const int maxBufCount = 20;
			const int bufSize = 0x400;
			PCacheString* cacheBuf = new PCacheString[maxBufCount];
			wchar_t stringsBuf[bufSize];
			int bufLen = bufSize;
			ExtractXmlValue(response, SEC_STRINGS_TAG, stringsBuf, &bufLen);
			if (bufLen > 0 && bufLen < bufSize) {
				DWORD cacheCount = 0;
				LPCWSTR stringPtr = stringsBuf;
				LPTSTR stringBuf = new TCHAR[bufSize];
				do {
					bufLen = bufSize;
					stringPtr = ExtractXmlValue(stringPtr, SEC_STRING_TAG, stringBuf, &bufLen);
					if (bufLen > 0) {
						// Extract the security String and Hash
						PCacheString newEntry = new CacheString();
						int strLen = 0x10;
						ExtractXmlValue(stringBuf, STRING_TAG, newEntry->securityString, &strLen);
						strLen = 0x40;
						ExtractXmlValue(stringBuf, HASH_TAG, newEntry->hash, &strLen);
						newEntry->used = 0;
						newEntry->deleted = 0;
						cacheBuf[cacheCount++] = newEntry;
					}
				} while (bufLen > 0);
				// Store the cached strings.
				CredentialCache* cache = CredentialCache::GetCacheForUser(username);
				cache->StoreCredentials(cacheBuf, cacheCount);
			}
		}
		delete response;
	}
	delete socket;
	delete request;
	return result;
}

BOOL PINsafeConnect::ChangePIN(LPCWSTR username, LPCWSTR oldPassword, LPCWSTR newPassword, LPCWSTR oldOtc, LPCWSTR newOtc) {
	BOOL result = FALSE;
	LPWSTR request = CreateChangePINRequest(username, oldPassword, newPassword, oldOtc, newOtc);
	HttpSocket* socket = new HttpSocket();
	socket->SetSecure(secure);
	socket->SetAllowSelfCert(allowSelfCert);
	LPWSTR response = socket->GetPage(agentUrl, TRUE, request);
	if (response != NULL) {
		wchar_t resBuf[0x10];
		int resLen = 0x10;
		ExtractXmlValue(response, RESULT_TAG, resBuf, &resLen);
		if (resLen > 0 && resLen < 0x10 && wcscmp(resBuf, RESULT_PASS) == 0) {
			result = TRUE;
		}
		delete response;
	}
	delete socket;
	delete request;
	return result;
}

IPicture* PINsafeConnect::GetSCImageBySessionId(LPCWSTR sessionId) {
	int urlLen = wcslen(scImageUrl) + wcslen(SESSION_ARG) + wcslen(sessionId) + 1;
	LPWSTR url = new wchar_t[urlLen];
	wcscpy(url, scImageUrl);
	wcscat(url, SESSION_ARG);
	wcscat(url, sessionId);
	return GetSCImage(url);
}

IPicture* PINsafeConnect::GetSCImageByUsername(LPCWSTR username) {
	int urlLen = wcslen(scImageUrl) + wcslen(USERNAME_ARG) + wcslen(username) + 1;
	LPWSTR url = new wchar_t[urlLen];
	wcscpy(url, scImageUrl);
	wcscat(url, USERNAME_ARG);
	wcscat(url, username);
	return GetSCImage(url);
}

void PINsafeConnect::RequestSecurityString(LPCWSTR username) {
	int urlLen = wcslen(dcImageUrl) + wcslen(USERNAME_ARG) + wcslen(username) + 1;
	LPWSTR url = new wchar_t[urlLen];
	wcscpy(url, dcImageUrl);
	wcscat(url, USERNAME_ARG);
	wcscat(url, username);
	IStream* strm = GetResponseStream(url);
	if (strm != NULL) {
		strm->Release();
	}
}

IStream* PINsafeConnect::GetResponseStream(LPCWSTR url) {
	HRESULT error;
	HttpSocket* socket = new HttpSocket();
	socket->SetSecure(secure);
	socket->SetAllowSelfCert(allowSelfCert);
	int cImgBytes;
	HGLOBAL imageBytes = (HGLOBAL)socket->GetPageBytes(url, FALSE, NULL, 0, &cImgBytes);
	if (imageBytes == NULL || cImgBytes == 0) {
		error = GetLastError();
		LogError(L"Error retrieving TURing image: %lx", error);
	} else {
		IStream* strm;
		error = CreateStreamOnHGlobal(imageBytes, TRUE, &strm);
		if (!error) {
			return strm;
		}
	}
	return NULL;
}

IPicture* PINsafeConnect::GetSCImage(LPCWSTR url) {
	IPicture* pTuring = NULL;
	HRESULT error;
	// Create an empty image
	error = OleCreatePictureIndirect(NULL, IID_IPicture, TRUE, (void**)&pTuring);
	if (!error) {
		IStream* strm = GetResponseStream(url);
		if (strm != NULL) {
			IPersistStream* pStrm;
			error = pTuring->QueryInterface(IID_IPersistStream, (LPVOID*)(&pStrm));
			if (!error) {
				pStrm->Load(strm);
				pStrm->Release();
				strm->Release();
			} 
		} else {
			error = E_FAIL;
		}
	}
//		error = OleLoadPicturePath((LPOLESTR)url, NULL, 0, NULL, IID_IPicture, (void**)&pTuring);
	if (error) {
		if (pTuring) {
			pTuring->Release();
			pTuring = NULL;
			LogError(L"Error retrieving TURing image");
		}
	}
	if (pTuring == NULL) {
		// TODO
		LogError(L"Error retrieving TURing image");
	}
	delete url;
	return pTuring;
}

BOOL PINsafeConnect::Ping() {
	int reqLen = 0x100;
	LPWSTR request = new wchar_t[reqLen];
	LPWSTR ptr = AppendStartTag(request, SAS_REQUEST_TAG, reqLen);
	ptr = AppendTagAndValue(ptr, ACTION_TAG, ACTION_PING, reqLen);
	AppendEndTag(ptr, SAS_REQUEST_TAG, reqLen);

	HttpSocket* socket = new HttpSocket();
	socket->SetSecure(secure);
	socket->SetAllowSelfCert(allowSelfCert);
	BOOL result = TRUE;
	LPWSTR response = socket->GetPage(agentUrl, TRUE, request);
	if (response == NULL) {
		result = FALSE;
	} else {
		// The response could be an error message - check for a result.
		const int RESULT_LEN = 10;
		wchar_t sResult[RESULT_LEN];
		int resLen = RESULT_LEN;
		ExtractXmlValue(response, RESULT_TAG, sResult, &resLen);
		if (resLen == 0) {
			result = FALSE;
		} else if (wcscmp(sResult, RESULT_PASS) != 0) {
			result = FALSE;
		}
	}

	delete socket;
	delete request;

	return result;
}

// Create a start session request with the buffer provided.
// Parameters:
//   request - the provided buffer.
//   username - the username to use.
//   secret - the secret to use.
//   reqLen - the length of the buffer.
// Returns the buffer if successful, or NULL if the buffer is not big enough.
LPWSTR TryCreateStartSessionRequest(LPWSTR request, LPCWSTR username, LPCWSTR secret, int reqLen) {
	LPWSTR ptr = AppendStartTag(request, SAS_REQUEST_TAG, reqLen);
	if (ptr == NULL) {
		return NULL;
	}
	ptr = AppendTagAndValue(ptr, ACTION_TAG, ACTION_STARTSESSION, reqLen);
	if (ptr == NULL) {
		return NULL;
	}
	ptr = AppendTagAndValue(ptr, USERNAME_TAG, username, reqLen);
	if (ptr == NULL) {
		return NULL;
	}
	ptr = AppendTagAndValue(ptr, SECRET_TAG, secret, reqLen);
	if (ptr == NULL) {
		return NULL;
	}
	ptr = AppendEndTag(ptr, SAS_REQUEST_TAG, reqLen);
	if (ptr == NULL) {
		return NULL;
	}

	return request;
}

// Create a start session request.
// Parameters:
//   username - the username to use.
// Returns a pointer to the request string.
LPWSTR PINsafeConnect::CreateStartSessionRequest(LPCWSTR username) {
	int reqLen = 0x200;
	LPWSTR request = new wchar_t[reqLen+1];
	while (TryCreateStartSessionRequest(request, username, secret, reqLen) == NULL) {
		delete request;
		reqLen += 0x80;
		request = new wchar_t[reqLen+1];
	}

	return request;
}

// Create a login request with the provided buffer.
// Parameters:
//   request - the request buffer.
//   username - the username to use.
//   password - the password to use.
//   otc - the one-time code to use.
//   secret - the secret to use.
//   reqLen - the size of the available buffer.
// Returns the buffer if successful, or NULL if the buffer is not big enough.
LPWSTR TryCreateLoginRequest(LPWSTR request, LPCWSTR username, LPCWSTR password, LPCWSTR otc, LPCWSTR secret, BOOL requestStrings, int reqLen) {
	LPWSTR ptr = AppendStartTag(request, SAS_REQUEST_TAG, reqLen);
	if (ptr == NULL) {
		return NULL;
	}
	ptr = AppendTagAndValue(ptr, ACTION_TAG, ACTION_LOGIN, reqLen);
	if (ptr == NULL) {
		return NULL;
	}
	ptr = AppendTagAndValue(ptr, SECRET_TAG, secret, reqLen);
	if (ptr == NULL) {
		return NULL;
	}
	ptr = AppendTagAndValue(ptr, USERNAME_TAG, username, reqLen);
	if (ptr == NULL) {
		return NULL;
	}
	ptr = AppendTagAndValue(ptr, PASSWORD_TAG, password, reqLen);
	if (ptr == NULL) {
		return NULL;
	}
	ptr = AppendTagAndValue(ptr, OTC_TAG, otc, reqLen);
	if (ptr == NULL) {
		return NULL;
	}
	if (requestStrings) {
		ptr = AppendStartTag(ptr, THIRD_PARTY_TAG, reqLen);
		if (ptr == NULL) {
			return NULL;
		}
		ptr = AppendTagAndValue(ptr, GINA_NAME_TAG, GINA_NAME_VALUE, reqLen);
		if (ptr == NULL) {
			return NULL;
		}
		RegSettings* settings = RegSettings::GetSettings();
		TCHAR cBuf[5];
		_itot_s(settings->GetCachedStringCount(), cBuf, 5, 10);
		ptr = AppendTagAndValue(ptr, GINA_DATA_TAG, cBuf, reqLen);
		if (ptr == NULL) {
			return NULL;
		}
		ptr = AppendEndTag(ptr, THIRD_PARTY_TAG, reqLen);
		if (ptr == NULL) {
			return NULL;
		}
	}
	ptr = AppendEndTag(ptr, SAS_REQUEST_TAG, reqLen);
	if (ptr == NULL) {
		return NULL;
	}

	return request;
}

// Create a login request.
// Parameters:
//   username - the username to use.
//   password - the password to use.
//   otc - the one-time code to use.
// Returns a pointer to the login request.
LPWSTR PINsafeConnect::CreateLoginRequest(LPCWSTR username, LPCWSTR password, LPCWSTR otc, BOOL requestStrings) {
	int reqLen = 0x200;
	LPWSTR request = new wchar_t[reqLen+1];
	while (TryCreateLoginRequest(request, username, password, otc, secret, requestStrings, reqLen) == NULL) {
		reqLen += 0x80;
		delete request;
		request = new wchar_t[reqLen+1];
	}

	return request;
}

// Create a change PIN request with the provided buffer.
// Parameters:
//   request - the request buffer.
//   username - the username to use.
//   oldPassword - the old password to use.
//   newPassword - the new password to use.
//   oldOtc - the old one-time code to use.
//   newOtc - the new one-time code to use.
//   secret - the secret to use.
//   reqLen - the size of the available buffer.
// Returns the buffer if successful, or NULL if the buffer is not big enough.
LPWSTR TryCreateChangePINRequest(LPWSTR request, LPCWSTR username, LPCWSTR oldPassword, LPCWSTR newPassword, LPCWSTR oldOtc, LPCWSTR newOtc, LPCWSTR secret, int reqLen) {
	LPWSTR ptr = AppendStartTag(request, SAS_REQUEST_TAG, reqLen);
	if (ptr == NULL) {
		return NULL;
	}
	ptr = AppendTagAndValue(ptr, ACTION_TAG, ACTION_CHANGEPIN, reqLen);
	if (ptr == NULL) {
		return NULL;
	}
	ptr = AppendTagAndValue(ptr, SECRET_TAG, secret, reqLen);
	if (ptr == NULL) {
		return NULL;
	}
	ptr = AppendTagAndValue(ptr, USERNAME_TAG, username, reqLen);
	if (ptr == NULL) {
		return NULL;
	}
	ptr = AppendTagAndValue(ptr, PASSWORD_TAG, oldPassword, reqLen);
	if (ptr == NULL) {
		return NULL;
	}
	ptr = AppendTagAndValue(ptr, NEWPASSWORD_TAG, newPassword, reqLen);
	if (ptr == NULL) {
		return NULL;
	}
	ptr = AppendTagAndValue(ptr, OTC_TAG, oldOtc, reqLen);
	if (ptr == NULL) {
		return NULL;
	}
	ptr = AppendTagAndValue(ptr, NEWOTC_TAG, newOtc, reqLen);
	if (ptr == NULL) {
		return NULL;
	}
	ptr = AppendEndTag(ptr, SAS_REQUEST_TAG, reqLen);
	if (ptr == NULL) {
		return NULL;
	}

	return request;
}

// Create a change PIN request.
// Parameters:
//   username - the username to use.
//   oldPassword - the old password to use.
//   newPassword - the new password to use.
//   oldOtc - the old one-time code to use.
//   newOtc - the new one-time code to use.
// Returns a pointer to the request string.
LPWSTR PINsafeConnect::CreateChangePINRequest(LPCWSTR username, LPCWSTR oldPassword, LPCWSTR newPassword, LPCWSTR oldOtc, LPCWSTR newOtc) {
	int reqLen = 0x200;
	LPWSTR request = new wchar_t[reqLen+1];
	while (TryCreateChangePINRequest(request, username, oldPassword, newPassword, oldOtc, newOtc, secret, reqLen) == NULL) {
		reqLen += 0x80;
		delete request;
		request = new wchar_t[reqLen+1];
	}

	return request;
}

BOOL PINsafeConnect::UserExists(LPCWSTR username) {
	BOOL exists = FALSE;
	LPWSTR request = CreateExistsRequest(username);
	HttpSocket* socket = new HttpSocket();
	socket->SetSecure(secure);
	socket->SetAllowSelfCert(allowSelfCert);
	LPWSTR response = socket->GetPage(agentUrl, TRUE, request);
	if (response != NULL) {
		wchar_t resBuf[0x10];
		int resLen = 0x10;
		ExtractXmlValue(response, RESULT_TAG, resBuf, &resLen);
		if (resLen > 0 && resLen < 0x10 && wcscmp(resBuf, RESULT_PASS) == 0) {
			exists = TRUE;
		}
		delete response;
	}
	delete socket;
	delete request;
	return exists;
}

LPWSTR TryCreateExistsRequest(LPWSTR request, LPCWSTR username, LPCWSTR secret, int reqLen) {
	LPWSTR ptr = AppendStartTag(request, SAS_REQUEST_TAG, reqLen);
	if (ptr == NULL) {
		return NULL;
	}
	ptr = AppendTagAndValue(ptr, ACTION_TAG, ACTION_EXISTS, reqLen);
	if (ptr == NULL) {
		return NULL;
	}
	ptr = AppendTagAndValue(ptr, USERNAME_TAG, username, reqLen);
	if (ptr == NULL) {
		return NULL;
	}
	ptr = AppendTagAndValue(ptr, SECRET_TAG, secret, reqLen);
	if (ptr == NULL) {
		return NULL;
	}
	ptr = AppendEndTag(ptr, SAS_REQUEST_TAG, reqLen);
	if (ptr == NULL) {
		return NULL;
	}

	return request;
}

LPWSTR PINsafeConnect::CreateExistsRequest(LPCWSTR username) {
	int reqLen = 0x200;
	LPWSTR request = new wchar_t[reqLen+1];
	while (TryCreateExistsRequest(request, username, secret, reqLen) == NULL) {
		delete request;
		reqLen += 0x80;
		request = new wchar_t[reqLen+1];
	}

	return request;
}