//
// THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY OF
// ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO
// THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A
// PARTICULAR PURPOSE.
//
// Copyright (c) 2006 Microsoft Corporation. All rights reserved.
//
// PINsafeLoginCredential is our implementation of ICredentialProviderCredential.
// ICredentialProviderCredential is what LogonUI uses to let a credential
// provider specify what a user tile looks like and then tell it what the
// user has entered into the tile.  ICredentialProviderCredential is also
// responsible for packaging up the users credentials into a buffer that
// LogonUI then sends on to LSA.

#pragma once

#include <windows.h>
#include <strsafe.h>
#include <shlguid.h>
#include "helpers.h"
#include "RegSettings.h"
#include "PINsafeLogin.h"
#include "TuringGenerator.h"
#include "PINsafeConnect.h"
#include "..\SwivelResource\SwivelResource.h"

class PINsafeLoginProvider;

class PINsafeLoginCredential : public ICredentialProviderCredential
{
    public:
    // IUnknown
    STDMETHOD_(ULONG, AddRef)()
    {
        return _cRef++;
    }
    
    STDMETHOD_(ULONG, Release)()
    {
        LONG cRef = _cRef--;
        if (!cRef)
        {
            delete this;
        }
        return cRef;
    }

    STDMETHOD (QueryInterface)(REFIID riid, void** ppv)
    {
        HRESULT hr;
        if (ppv != NULL)
        {
            if (IID_IUnknown == riid ||
                IID_ICredentialProviderCredential == riid)
            {
                *ppv = static_cast<IUnknown*>(this);
                reinterpret_cast<IUnknown*>(*ppv)->AddRef();
                hr = S_OK;
            }
            else
            {
                *ppv = NULL;
                hr = E_NOINTERFACE;
            }
        }
        else
        {
            hr = E_INVALIDARG;
        }
        return hr;
    }
  public:
    // ICredentialProviderCredential
    IFACEMETHODIMP Advise(ICredentialProviderCredentialEvents* pcpce);
    IFACEMETHODIMP UnAdvise();

    IFACEMETHODIMP SetSelected(BOOL* pbAutoLogon);
    IFACEMETHODIMP SetDeselected();

    IFACEMETHODIMP GetFieldState(DWORD dwFieldID,
                                 CREDENTIAL_PROVIDER_FIELD_STATE* pcpfs,
                                 CREDENTIAL_PROVIDER_FIELD_INTERACTIVE_STATE* pcpfis);

    IFACEMETHODIMP GetStringValue(DWORD dwFieldID, PWSTR* ppwsz);
    IFACEMETHODIMP GetBitmapValue(DWORD dwFieldID, HBITMAP* phbmp);
    IFACEMETHODIMP GetCheckboxValue(DWORD dwFieldID, BOOL* pbChecked, PWSTR* ppwszLabel);
    IFACEMETHODIMP GetComboBoxValueCount(DWORD dwFieldID, DWORD* pcItems, DWORD* pdwSelectedItem);
    IFACEMETHODIMP GetComboBoxValueAt(DWORD dwFieldID, DWORD dwItem, PWSTR* ppwszItem);
    IFACEMETHODIMP GetSubmitButtonValue(DWORD dwFieldID, DWORD* pdwAdjacentTo);

    IFACEMETHODIMP SetStringValue(DWORD dwFieldID, PCWSTR pwz);
    IFACEMETHODIMP SetCheckboxValue(DWORD dwFieldID, BOOL bChecked);
    IFACEMETHODIMP SetComboBoxSelectedValue(DWORD dwFieldID, DWORD dwSelectedItem);
    IFACEMETHODIMP CommandLinkClicked(DWORD dwFieldID);

    IFACEMETHODIMP GetSerialization(CREDENTIAL_PROVIDER_GET_SERIALIZATION_RESPONSE* pcpgsr, 
                                    CREDENTIAL_PROVIDER_CREDENTIAL_SERIALIZATION* pcpcs, 
                                    PWSTR* ppwszOptionalStatusText, 
                                    CREDENTIAL_PROVIDER_STATUS_ICON* pcpsiOptionalStatusIcon);
    IFACEMETHODIMP ReportResult(NTSTATUS ntsStatus, 
                                NTSTATUS ntsSubstatus,
                                PWSTR* ppwszOptionalStatusText, 
                                CREDENTIAL_PROVIDER_STATUS_ICON* pcpsiOptionalStatusIcon);

  public:
    HRESULT Initialize(const CREDENTIAL_PROVIDER_FIELD_DESCRIPTOR* rgcpfd,
                       const FIELD_STATE_PAIR* rgfsp,
                       ICredentialProviderCredential *pWrappedCredential,
                       DWORD dwWrappedDescriptorCount);
    PINsafeLoginCredential(PINsafeLoginProvider* provider);

    virtual ~PINsafeLoginCredential();

	void ShowTURingImageVisible();

  private:
    BOOL                _IsFieldInWrappedCredential(DWORD dwFieldID);
    FIELD_STATE_PAIR    *_LookupLocalFieldStatePair(DWORD dwFieldID);
	void				ShowPINsafeImage();
	IPicture*			GetTuringImage(PINsafeConnect* client);
	void				RequestSecurityString();
	BOOL				CheckUsername();
	int					GetFieldOffset(DWORD fieldID);
	void				ClearCredentials();
	BOOL				IsPasswordChangeRequired();
	void				CheckUsernameChanged();
	BOOL				UserExists(BOOL useLocal);
	BOOL				CheckUseLocal();
	void				ShowBusyCursor();			// Change the window cursor to the busy cursor
	void				ShowNormalCursor();			// Change the window cursor to the normal cursor

  private:
	RegSettings*							_settings; // PINsafe settings.
	PINsafeLoginProvider*					_provider; // The containing provider.
    LONG									_cRef;
    CREDENTIAL_PROVIDER_FIELD_DESCRIPTOR	_rgCredProvFieldDescriptors[PFI_NUM_FIELDS];// An array holding the type and name of each field in the tile.

    FIELD_STATE_PAIR						_rgFieldStatePairs[PFI_NUM_FIELDS];         // An array holding the state of each field in the tile.

    PWSTR									_rgFieldStrings[PFI_NUM_FIELDS];			// An array holding the string value of each field. This is 
                                                                                        // different from the name of the field held in 
                                                                                        // _rgCredProvFieldDescriptors.
	PWSTR									_pCredWaiting;				// String indicating waiting for PINsafe connection
	PWSTR									_pPINsafeUnavailable;		// String indicating that PINsafe is not available
	PWSTR									_pRequestImage;				// Request string text
	PWSTR									_pLinkText;					// Current text for request string link
    
    ICredentialProviderCredentialEvents		*_pCredProvCredentialEvents;                // Used to let our parent know when the credentials have changed.

    ICredentialProviderCredential        *_pWrappedCredential;                           // Our wrapped credential.
    DWORD                                _dwWrappedDescriptorCount;                      // The number of fields in our
                                                                                        // wrapped credential.
	DWORD	_turingBtnIndex;	// The index of the TURing button.

	ExpandableString*	_currentOTC;	// The OTC entered.
	ExpandableString*	_username;		// The username for the user attempting to log in.
	ExpandableString*	_newOTC;		// The new OTC for change PIN.
	ExpandableString*	_newOTC2;		// The confirmed new OTC for change PIN.
	ExpandableString*	_password;		// The current password.
	ExpandableString*	_newPassword;	// The new password (change password only).
	ExpandableString*	_newPassword2;	// Confirm new password.
	ExpandableString*	_domain;		// The current domain.
	ExpandableString*	_computerName;	// The computer name.
	ExpandableString*	_logonDomainInfo; // The prompt for the logon domain.
	BOOL	_credentialSelected;		// Flag indicating whether or not the credential is selected.
	BOOL	_useLocal;					// Flag indicating that local authentication is to be used.
	CTuringGenerator*	_turingLocal;	// Local TURing generator / authenticator.
	BOOL	_isChangePIN;				// TRUE if this credential is for Change PIN.
	BOOL	_isCredUI;					// TRUE if this credential is for Credential UI.
	HWND	_hCredWindow;				// The handle of the credential window.
	BOOL	_isShowingImage;			// TRUE if we are in the process of showing an image, to ignore additional clicks.
	BOOL	_isPINsafeAvailable;		// TRUE if PINsafe is available or status is unknown.
	BOOL	_PINChangeRequired;			// TRUE if the PIN must be changed.
	BOOL	_allowUnknownUsers;			// TRUE if unknown users can be authenticated without PINsafe.
	BOOL	_userExists;				// TRUE if the user exists, or unknown users are not permitted.
	ExpandableString*	_oldUsername;	// The last username checked.
};