//
// THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY OF
// ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO
// THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A
// PARTICULAR PURPOSE.
//
// Copyright (c) 2006 Microsoft Corporation. All rights reserved.
//
// PINsafeLoginProvider implements ICredentialProvider, which is the main
// interface that logonUI uses to decide which tiles to display.
// In this sample, we are wrapping the default password provider with
// an extra small text and combobox. We pass nearly all requests to the
// wrapped provider, except for the ones that are for fields we're
// responsible for ourselves. As far as the owner is concerned, we are a
// unique provider, so they never know we're wrapping another provider.

#include <credentialprovider.h>
#include "PINsafeLoginProvider.h"
#include "guid.h"
#include "PINsafeConnect.h"
#include "Logging.h"

// PINsafeLoginProvider ////////////////////////////////////////////////////////

PINsafeLoginProvider::PINsafeLoginProvider():
    _cRef(1)
{
    DllAddRef();

    _rgpCredentials = NULL;
    _dwCredentialCount = 0;

    _pWrappedProvider = NULL;
    _dwWrappedDescriptorCount = 0;

	_usernameIndex = -1;
	_passwordIndex = -1;
	_newPasswordIndex = -1;
	_confirmPasswordIndex = -1;
	_domainInfoIndex = -1;

	_fieldOffsets = NULL;
	_fieldIndices = NULL;

	_lastDomain = NULL;
	_lastUsername = NULL;
	_lastPassword = NULL;

	_settings = RegSettings::GetSettings();

	LPCWSTR resourceFile = _settings->GetResourceDll();
	_hResource = LoadLibraryEx(resourceFile, NULL, LOAD_LIBRARY_AS_DATAFILE);
	if (_hResource == NULL || _hResource == INVALID_HANDLE_VALUE) {
		DWORD err = GetLastError();
		LogError(L"Cannot open resource file %s: %x", resourceFile, err);
	}

	_passwordChangeRequired = FALSE;
	_passwordChangeComplete = FALSE;
	_isChangePIN = FALSE;
	_isCredUI = FALSE;
}

PINsafeLoginProvider::~PINsafeLoginProvider()
{
    _CleanUpAllCredentials();

	if (_fieldOffsets != NULL) {
		delete _fieldOffsets;
	}
	if (_lastDomain != NULL) {
		SecureZeroMemory(_lastDomain, wcslen(_lastDomain));
		CoTaskMemFree(_lastDomain);
	}
	if (_lastUsername != NULL) {
		SecureZeroMemory(_lastUsername, wcslen(_lastUsername));
		CoTaskMemFree(_lastUsername);
	}
	if (_lastPassword != NULL) {
		SecureZeroMemory(_lastPassword, wcslen(_lastPassword));
		CoTaskMemFree(_lastPassword);
	}
	FreeLibrary(_hResource);
    DllRelease();
}

// Cleans up all credentials, including the memory used to allocate the array.
void PINsafeLoginProvider::_CleanUpAllCredentials()
{
    // Iterate and clean up the array, if it exists.
    if (_rgpCredentials != NULL)
    {
        for (DWORD lcv = 0; lcv < _dwCredentialCount; lcv++)
        {
            if (_rgpCredentials[lcv] != NULL)
            {
                _rgpCredentials[lcv]->Release();
                _rgpCredentials[lcv] = NULL;
            }
        }
        delete [] _rgpCredentials;
        _rgpCredentials = NULL;
    }
}

// Ordinarily we would look at the CPUS and decide whether or not we support this scenario.
// However, in this scenario we're going to create our internal provider and let it answer
// questions like this for us.
HRESULT PINsafeLoginProvider::SetUsageScenario(
    CREDENTIAL_PROVIDER_USAGE_SCENARIO cpus,
    DWORD dwFlags
    )
{
    HRESULT hr;

	// Make sure these flags are turned off unless needed.
	_isChangePIN = FALSE;
	_isCredUI = FALSE;

    // Create the password credential provider and query its interface for an
    // ICredentialProvider we can use. Once it's up and running, ask it about the 
    // usage scenario being provided.
    IUnknown *pUnknown = NULL;
    hr = ::CoCreateInstance(CLSID_PasswordCredentialProvider, NULL, CLSCTX_ALL, IID_PPV_ARGS(&pUnknown));
    if (SUCCEEDED(hr))
    {
		hr = pUnknown->QueryInterface(IID_PPV_ARGS(&(_pWrappedProvider)));
		if (SUCCEEDED(hr))
		{
			hr = _pWrappedProvider->SetUsageScenario(cpus, dwFlags);
			if (SUCCEEDED(hr))
			{
				if (cpus == CPUS_CHANGE_PASSWORD) {
					// Change Password (PIN) is handled totally within this application.
					// All other scenarios use the wrapped provider.
					_isChangePIN = TRUE;
				}
				if (cpus == CPUS_CREDUI) {
					_isCredUI = true;
				}
			}
		}
    }
    if (FAILED(hr))
    {
		LogError(L"SetUsageScenario: unable to create wrapped instance");
        if (_pWrappedProvider != NULL)
        {
            _pWrappedProvider->Release();
            _pWrappedProvider = NULL;
        }
	}

    return hr;
}

// We pass this along to the wrapped provider.
STDMETHODIMP PINsafeLoginProvider::SetSerialization(
    const CREDENTIAL_PROVIDER_CREDENTIAL_SERIALIZATION* pcpcs
    )
{
    HRESULT hr = E_UNEXPECTED;
    
    if (_pWrappedProvider != NULL)
    {
        hr = _pWrappedProvider->SetSerialization(pcpcs);
    }

    return hr;
}

// Called by LogonUI to give you a callback. We pass this along to the wrapped provider.
HRESULT PINsafeLoginProvider::Advise(
    ICredentialProviderEvents* pcpe,
    UINT_PTR upAdviseContext
    )
{
    HRESULT hr = E_UNEXPECTED;
    if (_pWrappedProvider != NULL)
    {
        hr = _pWrappedProvider->Advise(pcpe, upAdviseContext);
    }
    return hr;
}

// Called by LogonUI when the ICredentialProviderEvents callback is no longer valid. 
// We pass this along to the wrapped provider.
HRESULT PINsafeLoginProvider::UnAdvise()
{
    HRESULT hr = E_UNEXPECTED;
    if (_pWrappedProvider != NULL)
    {
        hr = _pWrappedProvider->UnAdvise();
    }
    return hr;
}

// Called by LogonUI to determine the number of fields in your tiles.  This
// does mean that all your tiles must have the same number of fields.
// This number must include both visible and invisible fields. If you want a tile
// to have different fields from the other tiles you enumerate for a given usage
// scenario you must include them all in this count and then hide/show them as desired 
// using the field descriptors. We pass this along to the wrapped provider and then append
// our own credential count.
HRESULT PINsafeLoginProvider::GetFieldDescriptorCount(
    DWORD* pdwCount
    )
{
    HRESULT hr = E_UNEXPECTED;
	_pinsafeActive = TRUE;
	// For the sake of performance, we don't check for PINsafe availability until we need to.
/*	if (_isChangePIN) { // PINsafe must be active for Change PIN
		PINsafeConnect* client = new PINsafeConnect();
		if (!client->Ping()) {
			LogError(L"Unable to contact PINsafe server: cannot change PIN");
			*pdwCount = 0;
			hr = E_FAIL;
		}
		delete client;
	} else {*/
		// If we're using PINsafe for remote only, check if this is a remote session
		if (_settings->LoginDisabled()) 
		{
			_pinsafeActive = FALSE;
		}
		else if (!_settings->LoginLocal())
		{
			BOOL remote = GetSystemMetrics(SM_REMOTESESSION);
			if (!remote)
			{
				_pinsafeActive = FALSE;
			}
		}
		// If we're allowing fall-back, check if PINsafe is active.
/*		if (_pinsafeActive && _settings->AllowDefaultLogin())
		{
			PINsafeConnect* client = new PINsafeConnect();
			_pinsafeActive = client->Ping();
			delete client;
		}
	}*/

    if (_pWrappedProvider != NULL)
    {
        hr = _pWrappedProvider->GetFieldDescriptorCount(&(_dwWrappedDescriptorCount));
        if (SUCCEEDED(hr))
        {
			DWORD totalFields;
            // Note that we need to add our own credential count to the wrapped credential's
            // total count.
			// Initialise the credential offset maps.
			if (_fieldOffsets != NULL) {
				delete _fieldOffsets;
			}
			if (_fieldIndices != NULL) {
				delete _fieldIndices;
			}
			_fieldIndices = new int[PFI_NUM_FIELDS];
			for (int idx=0; idx<PFI_NUM_FIELDS; idx++) {
				_fieldIndices[idx] = -1;
			}
			if (_isChangePIN) {
				_fieldCount = PFCPI_NUM_FIELDS;
				_fieldOffsets = new DWORD[_fieldCount+2];
				for (int idx=0; idx<PFCPI_NUM_FIELDS; idx++) {
					_fieldOffsets[idx] = idx;
					_fieldIndices[idx] = idx;
				}
				if (_settings->ShowGetImage()) {
					_fieldIndices[PFI_TURING_BUTTON] = _fieldCount;
					_fieldOffsets[_fieldCount++] = PFI_TURING_BUTTON;
				}
				if (_settings->ShowRequestString()) {
					_fieldIndices[PFI_DEMAND_BUTTON] = _fieldCount;
					_fieldOffsets[_fieldCount++] = PFI_DEMAND_BUTTON;
				}
				totalFields = _fieldCount;
			} else {
				totalFields = _dwWrappedDescriptorCount;
				if (_pinsafeActive) {
					_fieldCount = 0;
					_fieldOffsets = new DWORD[PFI_USED_FIELDS];
					//_fieldOffsets[_fieldCount++] = PFI_NEW_PW;
					//_fieldOffsets[_fieldCount++] = PFI_CONFIRM_PW;
					_fieldIndices[PFI_OTC] = _fieldCount + _dwWrappedDescriptorCount;
					_fieldOffsets[_fieldCount++] = PFI_OTC;
					if (_settings->ShowGetImage()) {
						_fieldIndices[PFI_TURING_BUTTON] = _fieldCount + _dwWrappedDescriptorCount;
						_fieldOffsets[_fieldCount++] = PFI_TURING_BUTTON;
					}
					if (_settings->ShowRequestString()) {
						_fieldIndices[PFI_DEMAND_BUTTON] = _fieldCount + _dwWrappedDescriptorCount;
						_fieldOffsets[_fieldCount++] = PFI_DEMAND_BUTTON;
					}
					totalFields += _fieldCount;
				} else {
					_fieldOffsets = NULL;
				}
			}
			*pdwCount = totalFields;
        }
    }

    return hr;
}

LPCTSTR GuidToString(GUID guid)
{
	TCHAR buffer[40];
	_stprintf(buffer, 40, L"{%08x-%04x-%04x-%02x%02x-%02x%02x%02x%02x%02x%02x}", guid.Data1, guid.Data2, guid.Data3, guid.Data4[0], guid.Data4[1], guid.Data4[2], guid.Data4[3], 
		guid.Data4[4], guid.Data4[5], guid.Data4[6], guid.Data4[7]);

	return buffer;
}

// Gets the field descriptor for a particular field. If this descriptor refers to one owned
// by our wrapped provider, we'll pass it along. Otherwise we provide our own.
HRESULT PINsafeLoginProvider::GetFieldDescriptorAt(
    DWORD dwIndex, 
    CREDENTIAL_PROVIDER_FIELD_DESCRIPTOR** ppcpfd
    )
{
    HRESULT hr = E_UNEXPECTED;

	if (_isChangePIN) {
        if (ppcpfd != NULL) {
			int offset = GetFieldOffset(dwIndex);
			hr = FieldDescriptorCoAllocCopy(s_rgCredProvFieldDescriptors[offset], ppcpfd);
		} else {
            hr = E_INVALIDARG;
		}
	} else if (_pWrappedProvider != NULL) {
        if (ppcpfd != NULL)
        {
			if (dwIndex < _dwWrappedDescriptorCount) {
				hr = _pWrappedProvider->GetFieldDescriptorAt(dwIndex, ppcpfd);
				// Let's get some info from this field descriptor.
				if ((*ppcpfd)->cpft == CPFT_EDIT_TEXT && _usernameIndex < 0) {
					_usernameIndex = dwIndex;
				} else if ((*ppcpfd)->cpft == CPFT_PASSWORD_TEXT) {
					if (_passwordIndex < 0) {
						// We're assuming that the password field is the first password field found.
						_passwordIndex = dwIndex;
					} else if (_newPasswordIndex < 0) {
						_newPasswordIndex = dwIndex;
					} else if (_confirmPasswordIndex < 0) {
						_confirmPasswordIndex = dwIndex;
					}
				} else if ((*ppcpfd)->cpft == CPFT_SMALL_TEXT){
					if (_passwordIndex > 0 && (_domainInfoIndex < 0 || wcsicmp((*ppcpfd)->pszLabel, L"log on to") == 0)) {
						_domainInfoIndex = dwIndex;
					}
				}
			} else {
				DWORD offset = GetFieldOffset(dwIndex);
				if (offset < PFI_NUM_FIELDS) {
					hr = FieldDescriptorCoAllocCopy(s_rgCredProvFieldDescriptors[offset], ppcpfd);
				} else {
					hr = E_INVALIDARG;
				}
			}
        }
        else
        {
            hr = E_INVALIDARG;
        }
    }
	if (SUCCEEDED(hr)) {
		(*ppcpfd)->dwFieldID = dwIndex;
	}

    return hr;
}

// Sets pdwCount to the number of tiles that we wish to show at this time.
// Sets pdwDefault to the index of the tile which should be used as the default.
// The default tile is the tile which will be shown in the zoomed view by default. If 
// more than one provider specifies a default tile the last cred prov used can select
// the default tile. 
// If *pbAutoLogonWithDefault is TRUE, LogonUI will immediately call GetSerialization
// on the credential you've specified as the default and will submit that credential
// for authentication without showing any further UI.
// While we're here, we'll create credentials to wrap each of the credentials created by
// our wrapped provider. The key is to make everything transparent to the owner.
HRESULT PINsafeLoginProvider::GetCredentialCount(
    DWORD* pdwCount,
    DWORD* pdwDefault,
    BOOL* pbAutoLogonWithDefault
    )
{
    HRESULT hr = E_UNEXPECTED;
    DWORD dwDefault = 0;

	if (_hResource == NULL || _hResource == INVALID_HANDLE_VALUE) {
		// No valid resource: don't display credentials.
		return E_FAIL;
	}

	// Make sure we've created the provider.
    if (_pWrappedProvider != NULL)
    {
        // This probably shouldn't happen, but in the event that this gets called after
        // we've already been through once, we want to clean up everything before 
        // allocating new stuff all over again.
        if (_rgpCredentials != NULL)
        {
            _CleanUpAllCredentials();
        }

        // We need to know how many fields each credential has in order to initialize
        // our wrapper credentials, so we might as well do that here before anything else.
        DWORD count;
        hr = GetFieldDescriptorCount(&(count));

        if (SUCCEEDED(hr))
        {
			// Grab the credential count of the wrapped provider. We'll simply wrap each.
			// This is a dummy placeholder value: we never auto-logon.
			BOOL bAutoLogonWithDefault;
			hr = _pWrappedProvider->GetCredentialCount(&(_dwCredentialCount), &(dwDefault), &(bAutoLogonWithDefault));

            if (SUCCEEDED(hr))
            {
                // Create an array of credentials for use.
                _rgpCredentials = new PINsafeLoginCredential*[_dwCredentialCount];
                if (_rgpCredentials != NULL)
                {
                    // Iterate each credential and make a wrapper.
                    for (DWORD lcv = 0; SUCCEEDED(hr) && (lcv < _dwCredentialCount); lcv++)
                    {
                        // Allocate memory for the new credential.
                        _rgpCredentials[lcv] = new PINsafeLoginCredential(this);
                        if (_rgpCredentials[lcv] != NULL)
                        {
                            ICredentialProviderCredential *pCredential;
                            hr = _pWrappedProvider->GetCredentialAt(lcv, &(pCredential));
                            if (SUCCEEDED(hr))
                            {
                                // Set the Field State Pair and Field Descriptors for ppc's 
                                // fields to the defaults (s_rgCredProvFieldDescriptors, 
                                // and s_rgFieldStatePairs) and the value of SFI_USERNAME
                                // to pwzUsername.
								hr = _rgpCredentials[lcv]->Initialize(s_rgCredProvFieldDescriptors, s_rgFieldStatePairs, pCredential, _dwWrappedDescriptorCount);
                                if (FAILED(hr))
                                {
                                    // If initialization failed, clean everything up.
                                    if (pCredential != NULL)
                                    {
                                        pCredential->Release();
                                        pCredential = NULL;
                                    }

                                    for (lcv = 0; lcv < _dwCredentialCount; lcv++)
                                    {
                                        if (_rgpCredentials[lcv] != NULL)
                                        {
                                            // Release the pointer to account for the local reference.
                                            _rgpCredentials[lcv]->Release();
                                            _rgpCredentials[lcv] = NULL;
                                        }
                                    }
                                }

                            } // (End if _pWrappedProvider->GetCredentialAt succeeded.)
                        } // (End if allocating _rgpCredentials[lcv] succeeded.)
                        else
                        {
                            hr = E_OUTOFMEMORY;
                        } 
                    } // (End of _rgpCredentials allocation loop.)
                } // (End if for allocating _rgpCredentials succeeded.)
                else
                {
                    hr = E_OUTOFMEMORY;
                }
            } // (End if _pWrappedProvider->GetCredentialCount succeeded.)
        } // (End if GetFieldDescriptorCount succeeded.)
    }

    if (FAILED(hr))
    {
		LogError(L"GetCredentialCount: failed. Removing PINsafe credentials");
        // Clean up.
        if (_rgpCredentials != NULL)
        {
            delete _rgpCredentials;
            _rgpCredentials = NULL;
        }
    }
    else
    {
        *pdwCount = _dwCredentialCount;
        *pdwDefault = dwDefault;
        *pbAutoLogonWithDefault = FALSE;
    }

    return hr;
}

// Returns the credential at the index specified by dwIndex. This function is called by 
// logonUI to enumerate the tiles.
HRESULT PINsafeLoginProvider::GetCredentialAt(
    DWORD dwIndex, 
    ICredentialProviderCredential** ppcpc
    )
{
    HRESULT hr;

    // Validate parameters.
    if ((dwIndex < _dwCredentialCount) && 
        (ppcpc != NULL) &&
        (_rgpCredentials != NULL) &&
        (_rgpCredentials[dwIndex] != NULL))
    {
        hr = _rgpCredentials[dwIndex]->QueryInterface(IID_ICredentialProviderCredential, reinterpret_cast<void**>(ppcpc));
		if (FAILED(hr)) {
			LogError(L"Failed to get credential: error 0x%08X", hr);
		}
    }
    else
    {
		LogError(L"Invalid index, or credential is NULL");
        hr = E_INVALIDARG;
    }

    return hr;
}

// Retrieve a resource string.
LPCWSTR PINsafeLoginProvider::GetResourceString(UINT stringId) {
	const int BUFFER_SIZE = 0xff;
	wchar_t buffer[BUFFER_SIZE+1];
	LPWSTR result;
	int len = LoadString(_hResource, stringId, buffer, BUFFER_SIZE);
	if (len > 0) {
		SHStrDup(buffer, &result);
	} else {
		DWORD err = GetLastError();
		SHStrDup(L"", &result);
	}
	return result;
}

// Boilerplate code to create our provider.
HRESULT PINsafeLoginProvider_CreateInstance(REFIID riid, void** ppv)
{
    HRESULT hr;

    PINsafeLoginProvider* pProvider = new PINsafeLoginProvider();

    if (pProvider)
    {
        hr = pProvider->QueryInterface(riid, ppv);
        pProvider->Release();
    }
    else
    {
        hr = E_OUTOFMEMORY;
    }
    
    return hr;
}

void UnicodeStringToLPWSTR(LPCWSTR ustr, SIZE_T len, LPWSTR* str) {
	SIZE_T buflen = len+sizeof(WCHAR);
	LPVOID buffer = CoTaskMemAlloc(buflen);
	SecureZeroMemory(buffer, buflen);
	memcpy_s(buffer, buflen, ustr, len);
	*str = (LPWSTR)buffer;
}

void PINsafeLoginProvider::SaveLastCredentials(CREDENTIAL_PROVIDER_CREDENTIAL_SERIALIZATION *pcpcs, LPCWSTR password) {
	if (pcpcs != NULL) {
		KERB_INTERACTIVE_LOGON *pkil = (KERB_INTERACTIVE_LOGON*)pcpcs->rgbSerialization;
		if (pkil->MessageType == KerbInteractiveLogon) {
			if (_lastDomain != NULL) {
				CoTaskMemFree(_lastDomain);
			}
			if (_lastUsername != NULL) {
				CoTaskMemFree(_lastUsername);
			}
			UnicodeStringToLPWSTR((LPCWSTR)((BYTE*)pkil + (ULONG)pkil->LogonDomainName.Buffer), pkil->LogonDomainName.Length, &_lastDomain);
			UnicodeStringToLPWSTR((LPCWSTR)((BYTE*)pkil + (ULONG)pkil->UserName.Buffer), pkil->UserName.Length, &_lastUsername);
			//UnicodeStringToLPWSTR((LPCWSTR)((BYTE*)pkil + (ULONG)pkil->Password.Buffer), pkil->Password.Length, &_lastPassword);
		}
	}
	// For some reason (probably security), the serialized password is not correct. Use the passed-in value instead.
	if (password != NULL) {
		if (_lastPassword != NULL) {
			CoTaskMemFree(_lastPassword);
		}
		SHStrDup(password, &_lastPassword);
	}
}
