//
// THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY OF
// ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO
// THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A
// PARTICULAR PURPOSE.
//
// Copyright (c) Microsoft Corporation. All rights reserved.
//
//

#pragma once

#include <credentialprovider.h>
#include <windows.h>
#include <strsafe.h>

#include "PINsafeLoginCredential.h"
#include "helpers.h"
#include "RegSettings.h"

class PINsafeLoginProvider : public ICredentialProvider
{
  public:
    // IUnknown
    STDMETHOD_(ULONG, AddRef)()
    {
        return _cRef++;
    }
    
    STDMETHOD_(ULONG, Release)()
    {
        LONG cRef = _cRef--;
        if (!cRef)
        {
            delete this;
        }
        return cRef;
    }

    STDMETHOD (QueryInterface)(REFIID riid, void** ppv)
    {
        HRESULT hr;
        if (IID_IUnknown == riid || 
            IID_ICredentialProvider == riid)
        {
            *ppv = this;
            reinterpret_cast<IUnknown*>(*ppv)->AddRef();
            hr = S_OK;
        }
        else
        {
            *ppv = NULL;
            hr = E_NOINTERFACE;
        }
        return hr;
    }

  public:
    IFACEMETHODIMP SetUsageScenario(CREDENTIAL_PROVIDER_USAGE_SCENARIO cpus, DWORD dwFlags);
    IFACEMETHODIMP SetSerialization(const CREDENTIAL_PROVIDER_CREDENTIAL_SERIALIZATION* pcpcs);

    IFACEMETHODIMP Advise(__in ICredentialProviderEvents* pcpe, UINT_PTR upAdviseContext);
    IFACEMETHODIMP UnAdvise();

    IFACEMETHODIMP GetFieldDescriptorCount(__out DWORD* pdwCount);
    IFACEMETHODIMP GetFieldDescriptorAt(DWORD dwIndex,  __deref_out CREDENTIAL_PROVIDER_FIELD_DESCRIPTOR** ppcpfd);

    IFACEMETHODIMP GetCredentialCount(__out DWORD* pdwCount,
                                      __out DWORD* pdwDefault,
                                      __out BOOL* pbAutoLogonWithDefault);
    IFACEMETHODIMP GetCredentialAt(DWORD dwIndex, 
                                   __out ICredentialProviderCredential** ppcpc);

    friend HRESULT PINsafeLoginProvider_CreateInstance(REFIID riid, __deref_out void** ppv);

	int GetUsernameIndex() { return _usernameIndex; }
	int GetPasswordIndex() { return _passwordIndex; }
	int GetNewPasswordIndex() { return _newPasswordIndex; }
	int GetConfirmPasswordIndex() { return _confirmPasswordIndex; }
	int GetDomainInfoIndex() { return _domainInfoIndex; }
	bool IsChangePIN() { return _isChangePIN; }
	bool IsCredUI() { return _isCredUI; }
	bool IsPINsafeActive() { return _pinsafeActive; }
	int	GetFieldOffset(DWORD fieldID)
	{
		if (_isChangePIN) {
			if (fieldID >= _fieldCount) {
				return -0xff; // Invalid.
			}
			return _fieldOffsets[fieldID];
		} else {
			if (fieldID < _dwWrappedDescriptorCount) {
				return -1 - fieldID;
			}
			if (fieldID - _dwWrappedDescriptorCount >= _fieldCount) {
				return -0xff; // Invalid.
			}
			return _fieldOffsets[fieldID - _dwWrappedDescriptorCount];
		}
	}
	int GetFieldIndex(DWORD fieldOffset) {
		if (fieldOffset >= PFI_NUM_FIELDS) {
			return -1;
		}
		return _fieldIndices[fieldOffset];
	}
	DWORD GetAdditionalFieldCount() { return _fieldCount; }
	HINSTANCE GetResourceHandle() { return _hResource; }
	LPCWSTR GetResourceString(UINT stringId);
	void SetPasswordChangeRequired(BOOL required) { _passwordChangeRequired = required; }
	void SetPasswordChangeComplete(BOOL complete) { _passwordChangeRequired = FALSE; _passwordChangeComplete = complete; }
	BOOL IsPasswordChangeRequired() { return _passwordChangeRequired; }
	BOOL IsPasswordChangeComplete() { return _passwordChangeComplete; }
	void SetPINChangeRequired(BOOL required) { _PINChangeRequired = required; }
	void SaveLastCredentials(CREDENTIAL_PROVIDER_CREDENTIAL_SERIALIZATION *pcpcs, LPCWSTR password);
	LPCWSTR GetLastDomain() { return _lastDomain; }
	LPCWSTR GetLastUsername() { return _lastUsername; }
	LPCWSTR GetLastPassword() { return _lastPassword; }

protected:
    PINsafeLoginProvider();
    __override ~PINsafeLoginProvider();
    
private:
      void _CleanUpAllCredentials();
	  DWORD GetFieldCount();
    
private:
    LONG                _cRef;
    PINsafeLoginCredential   **_rgpCredentials;          // Pointers to the credentials which will be enumerated by this 
                                                    // Provider.

    ICredentialProvider *_pWrappedProvider;         // Our wrapped provider.
    DWORD               _dwCredentialCount;         // The number of credentials provided by our wrapped provider.
    DWORD               _dwWrappedDescriptorCount;  // The number of fields on each tile of our wrapped provider's 
                                                    // credentials.
    bool            _bEnumeratedSetSerialization;
	int				_usernameIndex; // The index of the username field.
	int				_passwordIndex; // The index of the password field.
	int				_newPasswordIndex; // The index of the new password field.
	int				_confirmPasswordIndex; // The index of the confirm new password field.
	int				_domainInfoIndex;	// The index of the string indicating which domain we are using.
	RegSettings*	_settings; // PINsafe settings.
	bool			_pinsafeActive; // Flag if PINsafe active.
	bool			_isChangePIN;
	bool			_isCredUI;
	DWORD*			_fieldOffsets;	// Array of offsets for displayed field into field lookup table.
	DWORD			_fieldCount;	// The number of local fields used.
	int*			_fieldIndices;	// Array of indices for our fields.
	BOOL	_passwordChangeRequired; // TRUE if the AD password must be changed.
	BOOL	_passwordChangeComplete; // TRUE if the AD password has just been changed.
	BOOL	_PINChangeRequired;	// TRUE if the PIN must be changed.

	HINSTANCE		_hResource;		// The resource DLL handle.

	LPWSTR			_lastUsername;	// The last username entered.
	LPWSTR			_lastPassword;	// The last password entered.
	LPWSTR			_lastDomain;	// The last domain name entered.
};