using System;
using System.IO;
using System.Net;
using System.Net.Security;
using System.Security.Cryptography.X509Certificates;
using System.Text;
using System.Windows.Forms;
using System.Xml;

namespace PINsafeLoginConfig
{
    public partial class SettingsForm : Form
    {
        protected LoginSettings settings;
        protected bool inLoad;
        
        public SettingsForm(LoginSettings settings)
        {
            InitializeComponent();

            this.settings = settings;
        }

        protected void SettingsFormLoad(object sender, EventArgs e)
        {
            // Set a flag so we know to ignore the value changed events emitted
            // when we load the current values
            inLoad = true;
            
            serverTextBox.Text = settings.Server;
            portUpDown.Value = Int32.Parse(settings.Port);
            contextTextBox.Text = settings.Context;
            secretTextBox.Text = confirmTextBox.Text = settings.Secret;

            secureCheckBox.Checked = selfSignedCheckBox.Enabled = settings.Secure;
            selfSignedCheckBox.Checked = settings.AllowSelfCert;

            switch (settings.LoginSelect)
            {
                case LoginSettings.PINSAFE_LOGIN_ALWAYS:
                    alwaysRadioButton.Checked = true;
                    break;
                case LoginSettings.PINSAFE_LOGIN_REMOTE:
                    remoteRadioButton.Checked = true;
                    break;
                default:
                    disabledRadioButton.Checked = true;
                    break;
            }

            rbUseStd.Checked = settings.AllowDefault;
        	rbFailAuth.Checked = (settings.UseLocal == LoginSettings.PINSAFE_LOCAL_NEVER && !settings.AllowDefault);
        	rbUseLocal.Checked = (settings.UseLocal == LoginSettings.PINSAFE_LOCAL_FALLBACK);
        	rbUseLocalAlways.Checked = (settings.UseLocal == LoginSettings.PINSAFE_LOCAL_ALWAYS);
            turingCheckBox.Checked = settings.ShowTuring;
			showRequestCheckBox.Checked = settings.ShowRequestString;
        	testModeCheckBox.Checked = settings.TestMode;
            ignoreDomainCheckBox.Checked = settings.IgnoreDomain;
            allowNonPINsafeCheckBox.Checked = settings.AllowUnknownUsers;

            if (serverTextBox.Text == "")
            {
                serverTextBox.Select();
            }

            // Try and make sure we actually get displayed
            Activate();
        }

        protected void SettingsFormActivated(object sender, EventArgs e)
        {
            // Now the form is loaded start paying attention to change events
            inLoad = false;
        }

        protected bool SaveSettings()
        {
            // Check secret confirmation
            if (secretTextBox.Modified)
            {
                if (secretTextBox.Text != confirmTextBox.Text)
                {
                    MessageBox.Show(this, "The entered secret and confirmation don't match", "Settings Error",
                                    MessageBoxButtons.OK, MessageBoxIcon.Stop);
                    secretTextBox.Clear();
                    confirmTextBox.Clear();
                    secretTextBox.Focus();

                    return false;
                }
            }

            // Save settings
            settings.Server = serverTextBox.Text;
            settings.Port = portUpDown.Value.ToString();
            settings.Context = contextTextBox.Text;
            settings.Secret = secretTextBox.Text;

            settings.Secure = secureCheckBox.Checked;
            settings.AllowSelfCert = selfSignedCheckBox.Checked;

            if (alwaysRadioButton.Checked)
            {
                settings.LoginSelect = LoginSettings.PINSAFE_LOGIN_ALWAYS;
            }
            else if (remoteRadioButton.Checked)
            {
                settings.LoginSelect = LoginSettings.PINSAFE_LOGIN_REMOTE;
            }
            else
            {
                settings.LoginSelect = LoginSettings.PINSAFE_LOGIN_DISABLED;
            }

            settings.AllowDefault = rbUseStd.Checked;
        	settings.UseLocal = (rbUseLocal.Checked)
        	                     	? LoginSettings.PINSAFE_LOCAL_FALLBACK
        	                     	: ((rbUseLocalAlways.Checked)
        	                     	   	? LoginSettings.PINSAFE_LOCAL_ALWAYS
        	                     	   	: LoginSettings.PINSAFE_LOCAL_NEVER);
            settings.ShowTuring = turingCheckBox.Checked;
			settings.ShowRequestString = showRequestCheckBox.Checked;
        	settings.TestMode = testModeCheckBox.Checked;
            settings.IgnoreDomain = ignoreDomainCheckBox.Checked;
            settings.AllowUnknownUsers = allowNonPINsafeCheckBox.Checked;

            settings.Save();

            return true;
        }

        protected void SecureCheckBoxCheckedChanged(object sender, EventArgs e)
        {
            if (!inLoad)
            {
                selfSignedCheckBox.Enabled = secureCheckBox.Checked;
            }
        }

        protected void ControlValueChanged(object sender, EventArgs e)
        {
            if (!inLoad)
            {
                applyButton.Enabled = true;
            }
        }

        protected void OkButtonClick(object sender, EventArgs e)
        {
            if (SaveSettings())
            {
                Close();
            }
            
            Close();
        }

        protected void ApplyButtonClick(object sender, EventArgs e)
        {
            if (SaveSettings())
            {
                // Disable apply button again until any changes are made
                applyButton.Enabled = false;
            }
        }

        protected void CancelButtonClick(object sender, EventArgs e)
        {
            Close();
        }

        private void HelpButtonClick(object sender, EventArgs e)
        {
            HelpForm helpForm = new HelpForm();
            helpForm.Show();
        }

		private bool DummyCertificateValidator(object sender, X509Certificate certificate, X509Chain chain, SslPolicyErrors sslPolicyErrors)
		{
			return true;
		}

		private bool StrictCertificateValidator(object sender, X509Certificate certificate, X509Chain chain, SslPolicyErrors sslPolicyErrors)
		{
			return (sslPolicyErrors == SslPolicyErrors.None);
		}

		/// Test the PINsafe connection
		private void BtnTestClick(object sender, EventArgs e)
		{
			StringBuilder url = new StringBuilder();
			url.Append(secureCheckBox.Checked ? "https://" : "http://");
			url.Append(serverTextBox.Text);
			url.Append(":");
			url.Append(portUpDown.Text);
			url.Append("/");
			url.Append(contextTextBox.Text);
			url.Append("/AgentXML");
			MessageBoxIcon icon = MessageBoxIcon.Error;
			String message;
			HttpWebRequest request = null;
			HttpWebResponse response;
			try
			{
				request = (HttpWebRequest) WebRequest.Create(url.ToString());
				request.Method = "POST";
				// Probably the best request for testing is "exists". We can ignore the result if it's a fail because the user doesn't exist.
				using (StreamWriter writer = new StreamWriter(request.GetRequestStream()))
				{
					writer.Write(
						"<?xml version=\"1.0\" encoding=\"UTF-8\"?><SASRequest><Action>exists</Action><Username>dummy</Username><Secret>");
					writer.Write(secretTextBox.Text);
					writer.Write("</Secret></SASRequest>");
				}
				if (selfSignedCheckBox.Checked)
				{
					// Enable dummy trust certificates.
					ServicePointManager.ServerCertificateValidationCallback = DummyCertificateValidator;
				}
				else
				{
					ServicePointManager.ServerCertificateValidationCallback = StrictCertificateValidator;
				}
				response = (HttpWebResponse)request.GetResponse();
				if (response.StatusCode == HttpStatusCode.OK)
				{
					// Check for result
					using (StreamReader reader = new StreamReader(response.GetResponseStream()))
					{
						string responseText = reader.ReadToEnd();
						try
						{
							XmlDocument responseDoc = new XmlDocument();
							responseDoc.LoadXml(responseText);
							XmlElement responseElem = responseDoc.DocumentElement;
							if (responseElem == null)
							{
								throw new XmlException("Empty document");
							}
							XmlElement resultElem = responseElem["Error"];
							if (resultElem != null)
							{
								// We've got an error. Check what it is.
								string error = resultElem.InnerText;
								if (!"".Equals(error))
								{
									icon = MessageBoxIcon.Warning;
									if ("AGENT_ERROR_UNAUTHORIZED".Equals(error))
									{
										message = "Either the PINsafe agent has not been defined, or the secret is wrong";
									}
									else
									{
										message = "The following error was received from PINsafe: \"" + error + "\"";
									}
								}
								else
								{
									message = "PINsafe Connection settings are correct";
									icon = MessageBoxIcon.Information;
								}
							}
							else
							{
								message = "PINsafe Connection settings are correct";
								icon = MessageBoxIcon.Information;
							}
						}
						catch (XmlException)
						{
							message = "The PINsafe server returned an invalid response";
						}
					}
				}
				else
				{
					switch (response.StatusCode)
					{
						case HttpStatusCode.NotFound:
							message = "The context \"" + contextTextBox.Text + "\" cannot be found";
							break;

						default:
							message = "The request returned an error status: \"" + response.StatusCode + ": " + response.StatusDescription +
									  "\"";
							break;
					}
				}
			}
			catch (WebException x)
			{
				WebExceptionStatus status = x.Status;
				switch (status)
				{
					case WebExceptionStatus.Timeout:
						message = "Unable to contact the PINsafe server";
						break;

					case WebExceptionStatus.NameResolutionFailure:
						message = "Unable to identify the server \"" + serverTextBox.Text + "\"";
						break;

					case WebExceptionStatus.TrustFailure:
						message =
							"The server certificate is not valid, and you have not checked the option to accept self-signed certificates";
						break;

					case WebExceptionStatus.ProtocolError:
						response = (HttpWebResponse) request.GetResponse();
						switch (response.StatusCode)
						{
							case HttpStatusCode.NotFound:
								message = "The context \"" + contextTextBox.Text + "\" cannot be found";
								break;

							default:
								message = "The request returned an error status: \"" + response.StatusCode + ": " + response.StatusDescription +
										  "\"";
								break;
						}
						break;

					default:
						message = "An error occurred connecting to the PINsafe server";
						break;
				}
			}
			catch (UriFormatException)
			{
				message = "The URL \"" + url + "\" is not valid";
			}
			catch (Exception x)
			{
				message = "An unexpected error occurred: " + x;
			}

			MessageBox.Show(message, "PINsafe connection Test", MessageBoxButtons.OK, icon);
		}

		private void ExportButton_Click(object sender, EventArgs e)
		{
			if (saveFileDialog.ShowDialog(this) == DialogResult.OK)
			{
				try
				{
					SaveSettings();
					settings.SaveToFile(saveFileDialog.FileName);

				}
				catch (Exception x)
				{
					MessageBox.Show("An error occurred trying to save the settings: " + x.Message, "PINsafe Login Configuration",
					                MessageBoxButtons.OK, MessageBoxIcon.Error);
				}
			}
		}

    }
}